#!/usr/bin/env python3
from argparse import ArgumentParser, Namespace, REMAINDER
from json import dumps
from os import environ
from subprocess import run
from textwrap import dedent

from dotenv import dotenv_values

from .identity import Identity, PolicyArn, Tag, DEFAULT_CLIENT


def main(help=False):
    cmd_funcs = {"whoami": whoami, "assume": assume_role, "exec": executor}

    parser = ArgumentParser(
        epilog="Switch roles, or through a chain or roles, or print identity information from AWS STS"
    )

    subparsers = parser.add_subparsers(dest="command")

    subparsers.add_parser(
        "whoami", epilog="Prints get-caller-identity info in JSON format"
    )

    assume = subparsers.add_parser(
        "assume",
        epilog=dedent(
            """
      Assume a role or a chain of roles with optional attributes, outputting the newly acquired credentials.
      Maintains parity with boto3's sts.assume_role except for MFA
    """
        ),
    )

    assume.add_argument(
        "-r",
        "--role-arn",
        help="""
      Role to assume. If declared multiple times each role will assume the next in the order given.
      All other options will be applied to all roles in the chain.
    """,
        action="append",
        required=True,
    )
    assume.add_argument(
        "-n",
        "--role-session-name",
        help="The session name to use with the role.",
        type=str,
        default="assumed-role",
    )
    assume.add_argument(
        "-p",
        "--policy-arn",
        help="Optional policy to attach to a session. Can be declared multiple times.",
        type=str,
        action="append",
    )
    assume.add_argument(
        "-t",
        "--tag",
        help="Optional tag to add to the session in the format of `mytagkey=myvalue`. Can be declared multiple times for multiple tags.",
        type=str,
        action="append",
    )
    assume.add_argument(
        "-T",
        "--transitive-tag-key",
        help="Transitive tag key. Can be declared multiple times.",
        type=str,
        action="append",
    )
    assume.add_argument(
        "-E",
        "--external-id",
        help="Optional External ID for the session. Required by some AssumeRole policies",
        type=str,
        default=None,
    )
    assume.add_argument(
        "-d",
        "--duration-seconds",
        help="Optional duration for the session.",
        type=int,
        default=3600,
    )

    # The order this comes in is imperative. It MUST come after all args in the parent
    # that we want but BEFORE those that we don't
    exec = subparsers.add_parser(
        "exec",
        epilog="Execute a command in a shell with newly created credentials.",
        parents=[assume],
        add_help=False
    )

    exec.add_argument(
       "-N",
       "--no-inherit-env",
       action="store_true",
       help="Don't allow the executed command to inherit the parent's env."
    )

    exec.add_argument(
       "-e",
       "--env-var",
       action="append",
       type=str,
       help="Env var in the format `MYVAR=foo` to pass to the executed command's environment. Can be declared multiple times."
    )

    exec.add_argument(
      "--env-file",
      type=str,
      help="Load env vars from a .env file."
    )

    assume.add_argument(
        "-e",
        "--env-vars",
        help="Output env vars usable from a terminal. If not set the output will match the output of aws-cli's `aws sts assume-role` JSON",
        action="store_true",
    )

    exec.add_argument(
        "exec_command",
        nargs=REMAINDER
    )

    if help:
      parser.print_help()
      return

    args = parser.parse_args()

    if args.command is None:
        parser.print_help()
        exit()

    return cmd_funcs[args.command](args)


def assume_role(args: Namespace, print_vars: bool = True):
    opts = dict(
        RoleArn=args.role_arn,
        RoleSessionName=args.role_session_name,
        PolicyArns=[PolicyArn(arn) for arn in (args.policy_arn or [])],
        Tags=[Tag(*pair) for pair in [tag.split("=") for tag in (args.tag or [])]],
        TransitiveTagKeys=args.transitive_tag_key or [],
    )

    if args.external_id:
        opts["ExternalId"] = args.external_id

    role = Identity(**opts)

    if print_vars:
      if args.env_vars:
          print(role.credentials.env_vars)
      else:
          res = dumps(role.credentials, indent=2)
          print(res)

    return role


def whoami(_: Namespace):
    res = DEFAULT_CLIENT.get_caller_identity()
    del res["ResponseMetadata"]
    print(dumps(res, indent=2))


def executor(args: Namespace):
  if not args.exec_command:
    print("Must supply a command to run.")
    main(help=True)
    exit()

  role = assume_role(args, print_vars=False)

  env = environ if not args.no_inherit_env else {}

  if args.env_var:
    env.update({
      pair[0]: pair[1] for pair in [ env_var.split("=") for env_var in args.env_var]
    })

  env.update({
    "AWS_ACCESS_KEY_ID": role.credentials.AccessKeyId,
    "AWS_SECRET_ACCESS_KEY": role.credentials.SecretAccessKey,
    "AWS_SESSION_TOKEN": role.credentials.SessionToken
  })

  if args.env_file:
    env.update(dotenv_values(args.env_file))

  run(
    args.exec_command,
    env=env,
    shell=False
  )


if __name__ == "__main__":
  main()


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aws_assumptions']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.26', 'python-dotenv>=1.0.0,<2.0.0']

entry_points = \
{'console_scripts': ['assumptions = aws_assumptions.scripts:main']}

setup_kwargs = {
    'name': 'aws-assumptions',
    'version': '0.2.0',
    'description': 'Assume role(s) from a terminal and easily manage boto3 clients for multiple identities at once.',
    'long_description': '# aws-assumptions\n\n* Easily switch between roles, or a chain of roles and create boto3 clients and resources off of those assumed identities.\n* Can be used as a library to assume roles. The created object also provides a factory for creating boto3 clients/resources off of the object\n* CLI script that allows printing credentials to stdout as either the standard response from boto3.sts.assume_role or formatted to use as env vars in a *nix terminal.\n* CLI provides `exec` command to execute terminal commands in a subshell with the newly minted credentials injected into the environment\n\n\n### Available CLI commands\n```\n~  > assumptions -h\nusage: assumptions [-h] {whoami,assume} ...\n\npositional arguments:\n  {whoami,assume,exec}\n\noptional arguments:\n  -h, --help       show this help message and exit\n\nSwitch roles, or through a chain or roles, or print identity information from AWS STS\n```\n\n### Getting current identity\n```\n> assumptions whoami -h\nusage: assumptions whoami [-h]\n\noptional arguments:\n  -h, --help  show this help message and exit\n\nPrints get-caller-identity info in JSON format\n```\n\n### Assuming a role\n```\n~  > assumptions assume -h\nusage: assumptions assume [-h] -r ROLE_ARN [-n ROLE_SESSION_NAME] [-p POLICY_ARN] [-t TAG] [-T TRANSITIVE_TAG_KEY] [-E EXTERNAL_ID] [-d DURATION_SECONDS] [-e]\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -r ROLE_ARN, --role-arn ROLE_ARN\n                        Role to assume. If declared multiple times each role will assume the next in the order given. All other options will be applied to all roles in the chain.\n  -n ROLE_SESSION_NAME, --role-session-name ROLE_SESSION_NAME\n                        The session name to use with the role.\n  -p POLICY_ARN, --policy-arn POLICY_ARN\n                        Optional policy to attach to a session. Can be declared multiple times.\n  -t TAG, --tag TAG     Optional tag to add to the session in the format of `mytagkey=myvalue`. Can be declared multiple times for multiple tags.\n  -T TRANSITIVE_TAG_KEY, --transitive-tag-key TRANSITIVE_TAG_KEY\n                        Transitive tag key. Can be declared multiple times.\n  -E EXTERNAL_ID, --external-id EXTERNAL_ID\n                        Optional External ID for the session. Required by some AssumeRole policies\n  -d DURATION_SECONDS, --duration-seconds DURATION_SECONDS\n                        Optional duration for the session.\n  -e, --env-vars        Output env vars usable from a terminal. If not set the output will match the output of aws-cli\'s `aws sts assume-role` JSON\n\nAssume a role or a chain of roles with optional attributes, outputting the newly acquired credentials. Maintains parity with boto3\'s sts.assume_role except for MFA\n```\n\nExample of assuming a role with env vars\n```\n> assumptions assume -r "arn:aws:iam::123456789876:role/my-role" -n bob@nowhere.com -e > creds.env\n> . creds.env\n```\n\nor\n\n```\n$(assumptions assume -r "arn:aws:iam::123456789876:role/my-role" -n bob@nowhere.com)\n```\n\n### Using `exec`\n\n```\nusage: scripts.py exec [-h] -r ROLE_ARN [-n ROLE_SESSION_NAME] [-p POLICY_ARN] [-t TAG] [-T TRANSITIVE_TAG_KEY] [-E EXTERNAL_ID] [-d DURATION_SECONDS] [-N] [-e ENV_VAR] [--env-file ENV_FILE] ...\n\npositional arguments:\n  exec_command\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -r ROLE_ARN, --role-arn ROLE_ARN\n                        Role to assume. If declared multiple times each role will assume the next in the order given. All other options will be applied to all roles in the chain.\n  -n ROLE_SESSION_NAME, --role-session-name ROLE_SESSION_NAME\n                        The session name to use with the role.\n  -p POLICY_ARN, --policy-arn POLICY_ARN\n                        Optional policy to attach to a session. Can be declared multiple times.\n  -t TAG, --tag TAG     Optional tag to add to the session in the format of `mytagkey=myvalue`. Can be declared multiple times for multiple tags.\n  -T TRANSITIVE_TAG_KEY, --transitive-tag-key TRANSITIVE_TAG_KEY\n                        Transitive tag key. Can be declared multiple times.\n  -E EXTERNAL_ID, --external-id EXTERNAL_ID\n                        Optional External ID for the session. Required by some AssumeRole policies\n  -d DURATION_SECONDS, --duration-seconds DURATION_SECONDS\n                        Optional duration for the session.\n  -N, --no-inherit-env  Don\'t allow the executed command to inherit the parent\'s env.\n  -e ENV_VAR, --env-var ENV_VAR\n                        Env var in the format `MYVAR=foo` to pass to the executed command\'s environment. Can be declared multiple times.\n  --env-file ENV_FILE   Load env vars from a .env file.\n\nExecute a command in a shell with newly created credentials.\n```\n\nExample\n```\n> assumptions exec -r "arn:aws:iam::123456789876:role/my-role" -n bob@nowhere.com aws sts get-caller-identity\n{\n    "UserId": "AROA4HO3IAI67GZHCWWWQ:bob@nowhere.com",\n    "Account": "840662778429",\n    "Arn": "arn:aws:sts::123456789876:assumed-role/my-role/bob@nowhere.com"\n}\n```\n\nExample passing env vars to an interactive shell\n```\n> assumptions exec -r "arn:aws:iam::123456789876:role/my-role" -n bob@nowhere.com -e FOO=bar bash\n$ echo $FOO\nbar\n```\n\n## Switching through multiple roles\nIf you need to chain roles (EG: Assume a role that assumes a role that assumes a role) you can pass the `-r` flag multiple times.\nNote however that all other options, such as `--external-id` or `--tag` will be applied to every session in the chain.\n\n## As a library\n\nAssuming a role and creating clients\n```python\nfrom aws_assumptions.identity import Identity\n\nsession = Identity(\n  RoleArn="arn:aws:iam::123456789876:role/my-role",\n  RoleSessionName="bob"\n)\n\nres = session.client("eks").list_clusters()\ncurrent_role = session.whoami()\nsession_that_made_current_rule = session.whomademe()\n```\n\nChaining roles\n\n```python\nfrom aws_assumptions.identity import Identity\n\nsession = Identity(\n  RoleArn=[\n    "arn:aws:iam::123456789876:role/my-role",\n    "arn:aws:iam::123456789876:role/my-second-role"\n  ],\n  RoleSessionName="bob"\n)\n\nres = session.client("eks").list_clusters()\ncurrent_role = session.whoami()\nsession_that_made_current_rule = session.whomademe()\n```\n\n',
    'author': 'Mathew Moon',
    'author_email': 'me@mathewmoon.net',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/mathewmoon/aws-assumptions',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
from argparse import ArgumentParser
from copy import deepcopy
from textwrap import dedent
from typing import Any, Dict, List, Union

from boto3.session import Session


SESSION = Session()
DEFAULT_CLIENT = Session().client("sts")


"""
The slottiness of the classes below are really just here so I get smacked in the face
with an error for bad attribute access but don't have to add a separate methods
to access as a dict. Really more of a debugging thing than anything else.
"""


class PolicyArn(dict):
    __slots__ = ["arn"]

    def __init__(self, arn):
        self.arn = arn
        return super().__init__(arn=arn)


class Tag(dict):
    __slots__ = ["Key", "Value"]

    def __init__(self, Key, Value):
        self.Key = Key
        self.Value = Value
        return super().__init__(Key=Key, Value=Value)


class AWSCredentials(dict):
    __slots__ = [
        "AccessKeyId",
        "SecretAccessKey",
        "SessionToken",
        "Expiration",
    ]

    def __getattribute__(self, __name: str) -> Any:
        # So I like dots..... Get over it.....
        if __name in super().__getattribute__("__slots__"):
            return super().__getitem__(__name)
        else:
            return super().__getattribute__(__name)

    @property
    def session_args(self):
        # Formatting matches that of boto3.session.Session, boto3.client, and boto3.resource
        # signatures. So helpful for creating new clients or sessions from creds off your object
        return {
            "aws_access_key_id": self.AccessKeyId,
            "aws_secret_access_key": self.SecretAccessKey,
            "aws_session_token": self.SessionToken,
        }

    @property
    def env_vars(self):
        # Useful when running from cli. eg: $(assume-role -r arn:iam::123456577:role/foo)
        # will get you using your new role from a terminal
        return dedent(
            f"""
      AWS_ACCESS_KEY_ID={self.AccessKeyId}
      AWS_SECRET_ACCESS_KEY={self.SecretAccessKey}
      AWS_SESSION_TOKEN={self.SessionToken}
    """
        )


class Identity:
    # Being slotty here to prevent abusing attributes by callers.
    # Just use the damn thing as intended.
    __slots__ = [
        "RoleArn",
        "PolicyArns",
        "RoleSessionName",
        "ExternalId",
        "TransitiveTagKeys",
        "Tags",
        "DurationSeconds",
        "boto_session",
        "__sts_client",
        "__credentials",
        "__opts",
        "__client_cache",
    ]

    def __init__(
        self,
        *,
        RoleArn: Union[List[str], str],
        DurationSeconds: int = 3600,
        ExternalId: str = None,
        PolicyArns: List[PolicyArn] = [],
        RoleSessionName: str = "aws-assumptions-session",
        TransitiveTagKeys: List[str] = [],
        Tags: List[Tag] = [],
        sts_client: object = None,
    ):
        opts = dict(
            RoleArn=RoleArn,
            PolicyArns=PolicyArns,
            RoleSessionName=RoleSessionName,
            ExternalId=ExternalId,
            TransitiveTagKeys=TransitiveTagKeys,
            Tags=Tags,
            DurationSeconds=DurationSeconds,
        )

        self.boto_session = SESSION
        self.__sts_client = sts_client or self.boto_session.client("sts")
        self.__credentials = None
        self.__client_cache = {"resource": {}, "client": {}}

        # Get rid of Nones and empties
        self.__opts = {k: v for k, v in opts.items() if v}

        # Handle jumping through lists of roles
        if isinstance(RoleArn, list):
            if not RoleArn:
                raise ValueError("RoleArn cannot be empty list")

            role_list = deepcopy(RoleArn)

            # For each ARN that we get we'll assume that identity
            # by using an STS client created off the prior object
            for arn in role_list:
                opts["RoleArn"] = arn
                new_identity = Identity(**opts)
                opts["sts_client"] = new_identity.client("sts")

            # Patch our own attributes til we look and quack like
            # a duck that hatched from the last ARN in the list
            self.RoleArn = arn
            self.__sts_client = opts["sts_client"]
            self.__opts["RoleArn"] = arn
            self.__credentials = new_identity.credentials
        else:
            self.__load_credentials()

    @property
    def credentials(self):
        if self.__credentials is None:
            self.__load_credentials()
        return self.__credentials

    def __load_credentials(self):
        client = self.__sts_client or self.boto_session.client("sts")
        res = client.assume_role(**self.__opts)
        self.__credentials = AWSCredentials(**res["Credentials"])

    def client(self, service: str, **kwargs: Dict[str, Any]):
        # A little factory for making boto3.client(s)
        opts = {**kwargs, **self.credentials.session_args}

        # If we don't have a cached client then make one
        if (
            not (client := self.__client_cache["client"].get(service))
            or client["opts"] != opts
        ):
            self.__client_cache["client"][service] = {
                "opts": opts,
                "client_obj": self.boto_session.client(service, **opts),
            }

        # Return client from cache
        return self.__client_cache["client"][service]["client_obj"]

    def resource(self, service: str, **kwargs: Dict[str, Any]):
        # A little factory for making boto3.resource(s)
        opts = {**kwargs, **self.credentials.session_args}

        # If we don't have a cached resource then make one
        if (
            not (client := self.__client_cache["resource"].get(service))
            or client["opts"] != opts
        ):
            self.__client_cache["service"][service] = {
                "opts": opts,
                "client_obj": self.boto_session.resource(service, **opts),
            }

        # Return client from cache
        return self.__client_cache["resource"][service]["client_obj"]

    def whoami(self):
        res = self.client("sts").get_caller_identity()
        del res["ResponseMetadata"]
        return res

    def whomademe(self):
        res = self.__sts_client.get_caller_identity()
        del res["ResponseMetadata"]
        return res
